from os.path import dirname

from pkglts.dependency import Dependency
from pkglts.option_object import Option
from pkglts.version import __version__


class OptionSphinx(Option):
    def version(self):
        return __version__

    def root_dir(self):
        return dirname(__file__)

    def update_parameters(self, cfg):
        sec = dict(
            theme="default",
            autodoc_dvlpt=True,
            build_dir="build/sphinx"
        )
        cfg['sphinx'] = sec

    def check(self, cfg):
        invalids = []
        if cfg['doc']['fmt'] != 'rst':
            invalids.append('doc.fmt')

        theme = cfg['sphinx']['theme']
        if theme != str(theme):
            invalids.append('sphinx.theme')

        return invalids

    def require_option(self):
        return ['test', 'doc', 'license']

    def require(self, cfg):
        yield Dependency('sphinx', intent='doc')
        if cfg["sphinx"]["theme"] == "sphinx_rtd_theme":
            yield Dependency('sphinx_rtd_theme', intent='doc')



from os import mkdir
from os.path import dirname, exists
from shutil import rmtree
from time import sleep


def ensure_created(dname):
    if not exists(dname):
        mkdir(dname)


def rmdir(dname):
    for i in range(5):
        if exists(dname):
            try:
                rmtree(dname)
                return
            except OSError:
                sleep(0.1)
        else:
            return

    raise OSError("unable to remove directory: %s" % dname)


def ensure_path(pth):
    dname = dirname(pth)
    if not exists(dname):
        ensure_path(dname)
        mkdir(dname)



from os.path import dirname

from pkglts.dependency import Dependency
from pkglts.option_object import Option
from pkglts.version import __version__


class OptionData(Option):
    def version(self):
        return __version__

    def root_dir(self):
        return dirname(__file__)

    def update_parameters(self, cfg):
        sec = dict(
            filetype=[".json", ".ini"],
            use_ext_dir=False
        )
        cfg['data'] = sec

    def require_option(self):
        return ['base']



import pytest

from pkglts.install_env.stdout_front_end import install, installed_packages

from .venv_tools import clear_venv, create_venv

__test__ = False


@pytest.fixture()
def venv():
    name = "tyti_stdout"
    mem = {}

    create_venv(venv, mem)

    yield name

    clear_venv(venv, mem)


def test_installed_packages():
    pkgs = set(installed_packages())
    assert "pkglts" in pkgs


def test_install(venv):
    assert "my-first-p" not in installed_packages()

    # install local wheel
    install("my-first-p")

    # test new package still not in virtualenv
    assert "my-first-p" not in installed_packages()



from os.path import dirname

from pkglts.dependency import Dependency
from pkglts.option.doc import fmt_badge
from pkglts.option_object import Option
from pkglts.version import __version__


class OptionCoveralls(Option):
    def version(self):
        return __version__

    def root_dir(self):
        return dirname(__file__)

    def require_option(self):
        return ['coverage', 'travis']

    def require(self, cfg):
        del cfg
        yield Dependency('coveralls', intent='test')

    def environment_extensions(self, cfg):
        owner = cfg['github']['owner']
        project = cfg['github']['project']

        url = "coveralls.io/github/%s/%s?branch=master" % (owner, project)
        img = ("coveralls.io/repos/github/%s/%s/" % (owner, project) +
               "badge.svg?branch=master")
        badge = fmt_badge(img, url, "Coverage report status", cfg['doc']['fmt'])

        return {"badge": badge}



"""
helper function for formatting documentation.
"""


def fmt_badge(badge, url, txt, fmt):
    """Produce valid img hyperlink.

    Args:
        badge (str): url of img
        url (str): url target of link
        txt (str): associated text
        fmt (str): doc format either 'rst' or 'md'

    Returns:
        (str)
    """
    if not url.startswith("http"):
        url = "https://%s" % url
    if not badge.startswith("http"):
        badge = "https://%s" % badge

    if fmt == 'rst':
        lines = [".. image:: %s" % badge,
                 "    :alt: %s" % txt,
                 "    :target: %s" % url]
        return "\n" + "\n".join(lines)

    if fmt == 'md':
        return "[![%s](%s)](%s)" % (txt, badge, url)

    raise UserWarning("Unknown format '{}'".format(fmt))



import pytest
from pkglts.config_management import Config
from pkglts.option.test.option import OptionTest


@pytest.fixture()
def opt():
    return OptionTest('test')


def test_root_dir_is_defined(opt):
    assert opt.root_dir() is not None


def test_update_parameters(opt):
    cfg = {}
    opt.update_parameters(cfg)
    assert len(cfg['test']) == 1


def test_config_check_suite_names(opt):
    for name in ('walou', ' nose'):
        cfg = Config(dict(test={'suite_name': name}))
        assert 'test.suite_name' in opt.check(cfg)


def test_require_option(opt):
    assert len(tuple(opt.require_option())) == 1


def test_require(opt):
    for suite_name in ('pytest', 'nose'):
        cfg = Config(dict(base={}, test={'suite_name': suite_name}))
        opt.update_parameters(cfg)

        assert len(tuple(opt.require(cfg))) == 2



from os import listdir
from os.path import join as pj
import pytest

from pkglts.manage import (get_pkg_config, get_pkg_hash,
                           init_pkg)

from .small_tools import ensure_created, rmdir


@pytest.fixture()
def tmp_dir():
    pth = 'toto_manage_cfg'
    ensure_created(pth)
    init_pkg(pth)

    yield pth

    rmdir(pth)


def test_manage_init_create_pkg_config(tmp_dir):
    # init_pkg(tmp_dir)
    cfg = get_pkg_config(tmp_dir)
    assert cfg is not None
    assert "_pkglts" in cfg


def test_manage_init_create_pkg_hash(tmp_dir):
    init_pkg(tmp_dir)
    hm = get_pkg_hash(tmp_dir)
    assert hm is not None


def test_manage_init_protect_pkglts_dir_from_modif(tmp_dir):
    assert "regenerate.no" in listdir(pj(tmp_dir, ".pkglts"))
    assert "clean.no" in listdir(pj(tmp_dir, ".pkglts"))


# @with_setup(setup, teardown)
# def test_manage_pkg_config():
#     cfg = dict(default_cfg)
#     cfg['toto'] = dict(toto=1)
#     write_pkg_config(cfg, tmp_dir)
#     new_cfg = get_pkg_config(tmp_dir)
#     assert new_cfg == cfg
#
#
# @with_setup(setup, teardown)
# def test_manage_pkg_config_fmt_templates():
#     cfg = dict(default_cfg)
#     cfg['base'] = dict(pkg='custom')
#     cfg['toto'] = dict(base=10, toto="{{key, base.pkg}}")
#     write_pkg_config(cfg, tmp_dir)
#
#     cfg = get_pkg_config(tmp_dir)['toto']
#     assert 'base' in cfg
#     assert 'toto' in cfg
#     assert cfg['toto'] == "custom"
#
#
# @with_setup(setup, teardown)
# def test_manage_cfg_store_any_item():
#     algo = sha512()
#     algo.update(("lorem ipsum\n" * 10).encode("latin1"))
#
#     cfg = dict(default_cfg)
#     cfg['toto'] = dict(simple=1,
#                        txt="lorem ipsum\n" * 4,
#                        hash=b64encode(algo.digest()).decode('utf-8'))
#
#     write_pkg_config(cfg, tmp_dir)
#
#     new_cfg = get_pkg_config(tmp_dir)
#     assert new_cfg == cfg
#
#     algo = sha512()
#     algo.update(("lorem ipsum\n" * 10).encode("latin1"))
#     sha = b64encode(algo.digest()).decode('utf-8')
#     assert sha == new_cfg['toto']['hash']
#
#
# @with_setup(setup, teardown)
# def test_manage_cfg_do_store_private_item():
#     cfg = dict(default_cfg)
#     cfg['toto'] = {}
#     cfg['_toto'] = {}
#     write_pkg_config(cfg, tmp_dir)
#
#     new_cfg = get_pkg_config(tmp_dir)
#     assert '_toto' in new_cfg
#
#
# @with_setup(setup, teardown)
# def test_manage_cfg_restore_templates_on_writing():
#     cfg = dict(default_cfg)
#     cfg['base'] = dict(pkg='custom')
#     cfg['toto'] = dict(base=10, toto="{{key, base.pkg}}")
#     write_pkg_config(cfg, tmp_dir)
#
#     pkg_cfg = get_pkg_config(tmp_dir)
#     assert pkg_cfg['toto']['toto'] == "custom"
#
#     pkg_cfg['base']['pkg'] = "another"
#     write_pkg_config(pkg_cfg, tmp_dir)
#
#     pkg_cfg = get_pkg_config(tmp_dir)
#     assert pkg_cfg['toto']['toto'] == "another"



from os.path import dirname, exists
from pkglts.option_object import Option
from pkglts.version import __version__

from . import nbcompile


class OptionNotebook(Option):
    def version(self):
        return __version__

    def root_dir(self):
        return dirname(__file__)

    def update_parameters(self, cfg):
        sec = dict(
            src_directory="example"
        )
        cfg['notebook'] = sec

    def check(self, cfg):
        invalids = []
        src_directory = cfg['notebook']['src_directory']

        if not exists(src_directory):
            invalids.append("notebook.src_directory")

        return invalids

    def require_option(self):
        return ['sphinx']

    def tools(self, cfg):
        del cfg
        yield nbcompile.parser_nbcompile



from os.path import dirname

from pkglts.option_object import Option
from pkglts.version import __version__


class OptionGitlab(Option):
    def version(self):
        return __version__

    def root_dir(self):
        return dirname(__file__)

    def update_parameters(self, cfg):
        sec = dict(
            owner="{{ base.authors[0][0] }}",
            project="{{ base.pkgname }}",
            server="framagit.org",
            url="https://{{ gitlab.server }}/{{ gitlab.owner }}/{{ gitlab.project }}"
        )
        cfg['gitlab'] = sec

    def check(self, cfg):
        invalids = []
        project = cfg['gitlab']['project']

        if not project:
            invalids.append('gitlab.project')

        return invalids

    def require_option(self):
        return ['git']



from os.path import dirname
from pkglts.option.pysetup.option import requirements
from pkglts.option_object import Option
from pkglts.version import __version__


class OptionConda(Option):
    def version(self):
        return __version__

    def root_dir(self):
        return dirname(__file__)

    def require_option(self):
        return ['pysetup']

    def environment_extensions(self, cfg):
        channels = set(dep.channel for dep in requirements(cfg))
        try:
            channels.remove(None)
        except KeyError:
            pass

        return {"channels": channels}



"""
This tool will try to bump the version number of the package.
"""
import logging

from pkglts.config_management import write_pkg_config

LOGGER = logging.getLogger(__name__)


def action_bump(cfg, **kwds):
    """Bump version number.
    """
    LOGGER.info("Bump version")
    pos = kwds['pos']
    # update both the actual config and its associated template
    for sec in (cfg['version'], cfg.template()['version']):
        if pos == 'major':
            sec['major'] += 1
            sec['minor'] = 0
            sec['post'] = 0
        elif pos == 'minor':
            sec['minor'] += 1
            sec['post'] = 0
        else:
            sec['post'] += 1

    write_pkg_config(cfg)


def parser_bump(subparsers):
    """Associate a CLI to this tool.

    Notes: The CLI will be a subcommand of pmg.

    Args:
        subparsers (ArgumentParser): entity to create a subparsers

    Returns:
        (string): a unique id for this parser
        (callable): the action to perform
    """
    parser = subparsers.add_parser('bump', help=action_bump.__doc__)
    parser.add_argument('pos', choices=['major', 'minor', 'post'],
                        help="Element of version to bump {major, minor, post}")

    return 'bump', action_bump



from os.path import dirname

from pkglts.option.doc import fmt_badge
from pkglts.option_object import Option
from pkglts.version import __version__


class OptionReadthedocs(Option):
    def version(self):
        return __version__

    def root_dir(self):
        return dirname(__file__)

    def update_parameters(self, cfg):
        sec = dict(
            project="{{ github.project }}"
        )
        cfg['readthedocs'] = sec

    def check(self, cfg):
        invalids = []
        project = cfg['readthedocs']['project']

        if not project:
            invalids.append("readthedocs.project")

        return invalids

    def require_option(self):
        return ['pysetup', 'github', 'sphinx']

    def environment_extensions(self, cfg):
        project = cfg['readthedocs']['project']
        project = project.replace(".", "")
        url = "%s.readthedocs.io/en/latest/?badge=latest" % project
        img = "readthedocs.org/projects/%s/badge/?version=latest" % project
        badge = fmt_badge(img, url, "Documentation status", cfg['doc']['fmt'])

        return {"badge": badge}



"""Conda front end to install package in current environment.
"""
from json import loads
from subprocess import Popen, PIPE


def installed_packages():
    """Iterate on all packages installed in the current python environment.

    return:
        (iter of str)
    """
    process = Popen(["conda", "list", "--json"], stdin=PIPE, stdout=PIPE, stderr=PIPE)
    output, _ = process.communicate("")
    if process.returncode != 0:
        raise UserWarning("unable to execute 'conda list'")

    for pkg in loads(output):
        yield pkg.split("-")[0]


def install(name):
    """Install a package in the current python environment.

    arg:
     - name (str): name of the package

    return:
     - (bool): whether installation was successful or not
    """
    process = Popen(["conda", "install", name], stdin=PIPE, stdout=PIPE, stderr=PIPE)
    process.communicate("")

    if process.returncode == 0:
        return True

    # try to use pip if conda install fail
    from .pip_front_end import install as pip_install
    return pip_install(name)



from os.path import dirname

from pkglts.dependency import Dependency
from pkglts.option.doc import fmt_badge
from pkglts.option_object import Option
from pkglts.version import __version__


class OptionAppveyor(Option):
    def version(self):
        return __version__

    def root_dir(self):
        return dirname(__file__)

    def update_parameters(self, cfg):
        sec = dict(
            token=""
        )
        cfg['appveyor'] = sec

    def require_option(self):
        return ['doc', 'pysetup', 'github']

    def environment_extensions(self, cfg):
        owner = cfg['github']['owner']
        project = cfg['github']['project'].replace("_", "-")
        token = cfg['appveyor']['token']

        url = "ci.appveyor.com/project/%s/%s/branch/master" % (owner, project)
        img = "ci.appveyor.com/api/projects/status/%s/branch/master?svg=true" % token
        badge = fmt_badge(img, url, "Appveyor build status", cfg['doc']['fmt'])

        return {"badge": badge}



import pytest
from pkglts.config_management import Config
from pkglts.option.sphinx.option import OptionSphinx


@pytest.fixture()
def opt():
    return OptionSphinx('sphinx')


def test_root_dir_is_defined(opt):
    assert opt.root_dir() is not None


def test_update_parameters(opt):
    cfg = {}
    opt.update_parameters(cfg)
    assert len(cfg['sphinx']) == 3


def test_config_checks_doc_fmt(opt):
    cfg = Config(dict(doc={'fmt': 'md'}, sphinx={'theme': "default"}))
    assert 'doc.fmt' in opt.check(cfg)


def test_config_check_sphinx_theme(opt):
    for theme in (1, None,):
        cfg = Config(dict(doc={'fmt': 'rst'}, sphinx={'theme': theme}))
        assert 'sphinx.theme' in opt.check(cfg)


def test_require_option(opt):
    assert len(tuple(opt.require_option())) == 3


def test_require(opt):
    cfg = Config()
    opt.update_parameters(cfg)

    assert len(tuple(opt.require(cfg))) == 1



import pytest
from pkglts.config_management import Config
from pkglts.option.appveyor.option import OptionAppveyor


@pytest.fixture()
def opt():
    return OptionAppveyor('appveyor')


def test_root_dir_is_defined(opt):
    assert opt.root_dir() is not None


def test_update_parameters(opt):
    cfg = {}
    opt.update_parameters(cfg)
    assert len(cfg['appveyor']) == 1


def test_config_does_nothing(opt):
    cfg = Config(dict(appveyor={'token': 'toto'}))
    assert len(opt.check(cfg)) == 0


def test_require_option(opt):
    assert len(tuple(opt.require_option())) == 3


def test_require(opt):
    cfg = Config(dict(base={}, travis={}))
    opt.update_parameters(cfg)

    assert len(tuple(opt.require(cfg))) == 0



import pytest
from pkglts.config_management import Config
from pkglts.option.pysetup.option import OptionPysetup


@pytest.fixture()
def opt():
    return OptionPysetup('pysetup')


def test_root_dir_is_defined(opt):
    assert opt.root_dir() is not None


def test_update_parameters(opt):
    cfg = {}
    opt.update_parameters(cfg)
    assert len(cfg['pysetup']) == 2


def test_config_check_intended_version_exists(opt):
    cfg = Config(dict(pysetup={'intended_versions': [], 'require': []}))
    assert 'pysetup.intended_versions' in opt.check(cfg)
    assert 'pysetup.require' not in opt.check(cfg)

    cfg = Config(dict(pysetup={'intended_versions': ["27"],
                               'require': [{'pkg_mng': 'walou', 'name': 'numpy'}]}))
    assert 'pysetup.require' in opt.check(cfg)


def test_require_option(opt):
    assert len(tuple(opt.require_option())) == 5


def test_require(opt):
    cfg = Config()
    opt.update_parameters(cfg)

    assert len(tuple(opt.require(cfg))) == 0



import pytest

from pkglts.install_env.load_front_end import get_install_front_end


def test_load_stdout_exists():
    ife = get_install_front_end('stdout')
    assert hasattr(ife, "install")


def test_load_pip_exists():
    ife = get_install_front_end('pip')
    assert hasattr(ife, "install")


def test_load_raise_import_error_if_not_available_name():
    with pytest.raises(ImportError):
        get_install_front_end("takapouet")



from os.path import dirname
from pkglts.dependency import Dependency
from pkglts.option_object import Option
from pkglts.option_tools import available_options
from pkglts.version import __version__

from . import find_requirements


class OptionPysetup(Option):
    def version(self):
        return __version__

    def root_dir(self):
        return dirname(__file__)

    def update_parameters(self, cfg):
        sec = dict(
            intended_versions=["36"],
            require=[]
        )
        cfg['pysetup'] = sec

    def check(self, cfg):
        invalids = []
        intended_versions = cfg['pysetup']['intended_versions']

        if not intended_versions:
            invalids.append("pysetup.intended_versions")

        req = cfg['pysetup']['require']

        valid_methods = (None, "pip", "conda", "git")
        if any(dep.get('pkg_mng') not in valid_methods for dep in req):
            invalids.append("pysetup.require")

        return invalids

    def require_option(self):
        return ['base', 'test', 'doc', 'license', 'version']

    def tools(self, cfg):
        del cfg
        yield find_requirements.parser_find_reqs

    def environment_extensions(self, cfg):
        reqs = requirements(cfg)

        def req(intent):
            """For internal use only."""
            return [r for r in reqs if r.intent == intent]

        def conda_reqs(intents):
            return fmt_conda_reqs(reqs, intents)

        def pip_reqs(intents):
            return fmt_pip_reqs(reqs, intents)

        cfg.add_test('is_pip_dep', Dependency.is_pip)

        return {"pkg_url": pkg_url(cfg),
                "requirements": req,
                "conda_reqs": conda_reqs,
                "pip_reqs": pip_reqs}


def requirements(cfg):
    """Check all requirements for installed options.

    Args:
        cfg (Config):  current package configuration

    Returns:
        (list): list of required packages
    """
    reqs = {}
    for name in cfg.installed_options():
        try:
            opt = available_options[name]
            for dep in opt.require(cfg):
                reqs[dep.name] = dep
        except KeyError:
            raise KeyError("option '%s' does not exists" % name)

    for dep_def in cfg['pysetup']['require']:
        dep = Dependency(**dep_def)
        reqs[dep.name] = dep

    return [reqs[name] for name in sorted(reqs)]


def pkg_url(cfg):
    """Extract a valid url from all config sections.

    Args:
        cfg (Config):  current package configuration

    Returns:
        (str): a valid url for the package
    """
    try:
        url = cfg['base']['url']
        if url is not None:
            return url
    except KeyError:
        pass

    try:
        url = cfg['github']['url']
        if url is not None:
            return url
    except KeyError:
        pass

    try:
        url = cfg['gitlab']['url']
        if url is not None:
            return url
    except KeyError:
        pass

    # try:
    #     url = cfg['pypi']['url']
    #     if url is not None:
    #         return url
    # except KeyError:
    #     pass

    # try:
    #     url = cfg['readthedocs']['url']
    #     if url is not None:
    #         return url
    # except KeyError:
    #     pass

    return ""


def fmt_conda_reqs(reqs, intents):
    """Produce conda cmd line to install list of requirements.

    Args:
        reqs (list of Dependency): list of requirements objects
        intents (list of str): list of intents for deps

    Returns:
        (str)
    """
    reqs = [r for r in reqs if r.is_conda(strict=False) and r.intent in intents]
    if len(reqs) == 0:
        return ""

    cmd = "conda install"
    for channel in set(r.channel for r in reqs) - {None}:
        cmd += " -c %s" % channel

    for name in sorted(r.conda_full_name() for r in reqs):
        cmd += " %s" % name

    return cmd


def fmt_pip_reqs(reqs, intents):
    """Produce pip cmd line to install list of requirements.

    Args:
        reqs (list of Dependency): list of requirements objects
        intents (list of str): list of intents for deps

    Returns:
        (str)
    """
    reqs = [r for r in reqs if r.is_pip(strict=True) and r.intent in intents]
    if len(reqs) == 0:
        return ""

    cmd = "pip install"

    for name in sorted(r.pip_full_name() for r in reqs):
        cmd += " %s" % name

    return cmd



from copy import deepcopy
from os import listdir
from os.path import exists
from os.path import join as pj
import pytest

from pkglts.config_management import Config, DEFAULT_CFG
from pkglts.manage import install_example_files

from .small_tools import ensure_created, rmdir


@pytest.fixture()
def tmp_dir():
    pth = "tmp_cfgex"
    ensure_created(pth)

    yield pth

    rmdir(pth)


def test_install_example_returns_false_if_option_not_already_installed(tmp_dir):
    cfg = Config(DEFAULT_CFG)
    cfg.load_extra()
    ans = install_example_files('option', cfg, tmp_dir)
    assert not ans


def test_install_example_ok_if_option_do_not_provide_examples(tmp_dir):
    pkg_cfg = deepcopy(DEFAULT_CFG)
    pkg_cfg['base'] = dict(pkgname='toto', namespace=None)
    cfg = Config(pkg_cfg)
    cfg.load_extra()

    ans = install_example_files('base', cfg, tmp_dir)
    assert not ans


def test_install_example_copy_files(tmp_dir):
    pkg_cfg = deepcopy(DEFAULT_CFG)
    pkg_cfg['base'] = dict(pkgname='toto', namespace=None)
    pkg_cfg['test'] = dict(suite_name='nose')
    cfg = Config(pkg_cfg)
    cfg.load_extra()

    assert len(listdir(tmp_dir)) == 0
    install_example_files('test', cfg, tmp_dir)
    assert len(listdir(tmp_dir)) > 0
    assert exists(pj(tmp_dir, "src", "toto", "example.py"))
    assert exists(pj(tmp_dir, "test", "test_example.py"))


def test_install_example_copy_binary_files(tmp_dir):
    pkg_cfg = deepcopy(DEFAULT_CFG)
    pkg_cfg['base'] = dict(pkgname='toto', namespace=None)
    pkg_cfg['data'] = dict()
    cfg = Config(pkg_cfg)
    cfg.load_extra()

    assert len(listdir(tmp_dir)) == 0
    install_example_files('data', cfg, tmp_dir)
    assert len(listdir(tmp_dir)) > 0
    assert exists(pj(tmp_dir, "src", "toto_data", "ext_data.png"))


def test_install_example_do_not_complain_if_file_already_exists(tmp_dir):
    pkg_cfg = deepcopy(DEFAULT_CFG)
    pkg_cfg['base'] = dict(pkgname='toto', namespace=None)
    pkg_cfg['test'] = dict(suite_name='nose')
    cfg = Config(pkg_cfg)
    cfg.load_extra()

    assert install_example_files('test', cfg, tmp_dir)
    assert install_example_files('test', cfg, tmp_dir)


def test_install_example_handles_namespace(tmp_dir):
    pkg_cfg = deepcopy(DEFAULT_CFG)
    pkg_cfg['base'] = dict(pkgname='toto', namespace='oa', namespace_method='pkg_utils')
    pkg_cfg['test'] = dict(suite_name='nose')
    cfg = Config(pkg_cfg)
    cfg.load_extra()

    install_example_files('test', cfg, tmp_dir)
    assert exists(pj(tmp_dir, "test", "test_example.py"))
    assert not exists(pj(tmp_dir, "src", "toto", "example.py"))
    assert exists(pj(tmp_dir, "src", "oa", "toto", "example.py"))



from pkglts.config_management import Config


def test_contributors_with_failed_request(mocker):
    with mocker.patch('subprocess.check_output', side_effect=KeyError):
        cfg = Config(dict(git={}))
        cfg.load_extra()
        assert "failed" in cfg._env.globals['git'].contributors[0]


def test_contributors_without_failed_request(mocker):
    commits = b'commit Author: a <toto@titi>\n, commit: Author: b <tata@titi>\n, commit: Author: b <tata@titi>\n'
    with mocker.patch('subprocess.check_output', return_value=commits):
        cfg = Config(dict(git={}))
        cfg.load_extra()
        contributors = cfg._env.globals['git'].contributors
        assert len(contributors) == 2



from os.path import dirname

from pkglts.dependency import Dependency
from pkglts.option_object import Option
from pkglts.version import __version__


class OptionCoverage(Option):
    def version(self):
        return __version__

    def root_dir(self):
        return dirname(__file__)

    def require_option(self):
        return ['test']

    def require(self, cfg):
        yield Dependency('coverage', intent='test')
        if cfg['test']['suite_name'] == 'pytest':
            yield Dependency('pytest-cov', intent='test')



from os.path import join as pj
from os.path import exists
import pytest
from random import random

from pkglts.config_management import Config, DEFAULT_CFG
from pkglts.hash_management import pth_as_key
from pkglts.manage_tools import regenerate_dir

from .small_tools import ensure_created, ensure_path, rmdir


@pytest.fixture()
def tmp_dir():
    pth = "takapouet"

    ensure_created(pth)
    ensure_created(pj(pth, "src"))
    ensure_created(pj(pth, "tgt"))

    yield pth

    rmdir(pth)


def test_regenerate_dir_walk_all_files_in_src_dir(tmp_dir):
    ensure_created(pj(tmp_dir, "src", "sub"))
    fnames = ('toto.txt', 'titi.txt', 'sub/toto.txt')
    for fname in fnames:
        pth = pj(tmp_dir, "src", fname)
        with open(pth, 'w') as f:
            f.write("lorem ipsum")

    cfg = Config(DEFAULT_CFG)
    regenerate_dir(pj(tmp_dir, 'src'), pj(tmp_dir, 'tgt'), cfg, {})

    for fname in fnames:
        assert exists(pj(tmp_dir, 'tgt', fname))


def test_regenerate_dir_render_file_content(tmp_dir):
    pth = pj(tmp_dir, "src", "test.txt")
    with open(pth, 'w') as f:
        f.write("{{ 'lorem ipsum'|upper }}")

    cfg = Config(DEFAULT_CFG)
    regenerate_dir(pj(tmp_dir, 'src'), pj(tmp_dir, 'tgt'), cfg, {})

    pth = pj(tmp_dir, "tgt", "test.txt")
    with open(pth, 'r') as f:
        cnt = f.read()
        assert cnt == 'LOREM IPSUM'


def test_regenerate_dir_render_path_names(tmp_dir):
    ensure_created(pj(tmp_dir, "src", "{{ custom_name }}"))
    fnames = ('{{ custom_name }}.txt', 'titi.txt',
              '{{ custom_name }}/{{ custom_name }}.txt')
    for fname in fnames:
        pth = pj(tmp_dir, "src", fname)
        with open(pth, 'w') as f:
            f.write("lorem ipsum")

    cfg = Config(DEFAULT_CFG)
    cfg._env.globals['custom_name'] = 'toto'
    regenerate_dir(pj(tmp_dir, 'src'), pj(tmp_dir, 'tgt'), cfg, {})

    fnames = ('toto.txt', 'titi.txt', 'toto/toto.txt')
    for fname in fnames:
        assert exists(pj(tmp_dir, 'tgt', fname))


def test_regenerate_handle_src_directory_no_namespace(tmp_dir):
    pth = pj(tmp_dir, "src", "src", "{{ base.pkgname }}", "test.txt")
    ensure_path(pth)
    with open(pth, 'w') as f:
        f.write("lorem ipsum")

    pkg_cfg = dict(DEFAULT_CFG)
    pkg_cfg['base'] = dict(pkgname='toto', namespace=None)
    cfg = Config(pkg_cfg)
    regenerate_dir(pj(tmp_dir, 'src'), pj(tmp_dir, 'tgt'), cfg, {})

    tgt = pj(tmp_dir, "tgt", "src", "toto", "test.txt")
    assert exists(tgt)


def test_regenerate_handle_src_directory_with_namespace(tmp_dir):
    pth = pj(tmp_dir, "src", "src", "{{ base.pkgname }}", "test.txt")
    ensure_path(pth)
    with open(pth, 'w') as f:
        f.write("lorem ipsum")

    pkg_cfg = dict(DEFAULT_CFG)
    pkg_cfg['base'] = dict(pkgname='toto', namespace='myns', namespace_method='pkg_utils')
    cfg = Config(pkg_cfg)
    regenerate_dir(pj(tmp_dir, 'src'), pj(tmp_dir, 'tgt'), cfg, {})

    tgt_dir = pj(tmp_dir, "tgt")
    assert exists(tgt_dir + "/src")
    assert exists(tgt_dir + "/src/myns")
    assert exists(tgt_dir + "/src/myns/__init__.py")
    assert exists(tgt_dir + "/src/myns/toto")
    assert exists(tgt_dir + "/src/myns/toto/test.txt")


def test_regenerate_do_overwrite_unprotected_files(tmp_dir):
    pth = pj(tmp_dir, "src", "test.txt")
    with open(pth, 'w') as f:
        f.write("{# pkglts, b0\n{{ random() }}\n#}\n")

    cfg = Config(DEFAULT_CFG)
    cfg._env.globals['random'] = random
    regenerate_dir(pj(tmp_dir, 'src'), pj(tmp_dir, 'tgt'), cfg, {})

    pth = pj(tmp_dir, "tgt", "test.txt")
    with open(pth, 'r') as f:
        cnt0 = f.read()

    regenerate_dir(pj(tmp_dir, 'src'), pj(tmp_dir, 'tgt'), cfg, {})

    pth = pj(tmp_dir, "tgt", "test.txt")
    with open(pth, 'r') as f:
        cnt1 = f.read()

    assert cnt0 != cnt1


def test_regenerate_do_not_overwrite_protected_files(tmp_dir):
    pth = pj(tmp_dir, "src", "test.txt")
    with open(pth, 'w') as f:
        f.write("{# pkglts, b0\n{{ random() }}\n#}\n")

    cfg = Config(DEFAULT_CFG)
    cfg._env.globals['random'] = random
    regenerate_dir(pj(tmp_dir, 'src'), pj(tmp_dir, 'tgt'), cfg, {})

    pth = pj(tmp_dir, "tgt", "test.txt")
    with open(pth, 'r') as f:
        cnt0 = f.read()

    overwrite = {pth_as_key(pth): False}
    regenerate_dir(pj(tmp_dir, 'src'), pj(tmp_dir, 'tgt'), cfg, overwrite)

    pth = pj(tmp_dir, "tgt", "test.txt")
    with open(pth, 'r') as f:
        cnt1 = f.read()

    assert cnt0 == cnt1


def test_regenerate_do_not_overwrite_outside_protected_blocks(tmp_dir):
    pth = pj(tmp_dir, "src", "test.txt")
    with open(pth, 'w') as f:
        f.write("{# pkglts, b0\nLOREM IPSUM\n#}\n")

    pth = pj(tmp_dir, "tgt", "test.txt")
    with open(pth, 'w') as f:
        f.write("Toto start\n{# pkglts, b0\nWTF?\n#}\nToto end\n")

    cfg = Config(DEFAULT_CFG)
    regenerate_dir(pj(tmp_dir, 'src'), pj(tmp_dir, 'tgt'), cfg, {})

    with open(pth, 'r') as f:
        cnt = f.read()

    assert cnt == "Toto start\n{# pkglts, b0\nLOREM IPSUM\n#}\nToto end\n"




"""Factory to load a specific front end.
"""
from importlib import import_module


def get_install_front_end(name):
    """Load install front end.

    args:
     - name (str): name of front end to load

    return:
     - (module): loaded python module
    """
    try:
        return import_module("pkglts.install_env.%s_front_end" % name)
    except ImportError:
        raise ImportError("Install front end '%s' does not exist" % name)



""" Contains functions to manage the structure of the package.

Use 'setup.py' for common tasks.
"""
import logging
from os import listdir, mkdir, remove, walk
from os.path import exists, normpath, splitext
from os.path import join as pj
from shutil import rmtree

from .config import pkg_cfg_file, pkg_hash_file, pkglts_dir
from .config_management import (Config, DEFAULT_CFG,
                                get_pkg_config, write_pkg_config)
from .hash_management import (get_pkg_hash, modified_file_hash,
                              pth_as_key, write_pkg_hash)
from .manage_tools import (check_option_parameters,
                           regenerate_dir, update_opt)
from .option_tools import available_options, get_user_permission

LOGGER = logging.getLogger(__name__)


def init_pkg(rep="."):
    """Initialise a package in given directory.

    Args:
        rep (str): directory to create pkg into, default current

    Returns:
        None
    """
    if not exists(pj(rep, pkglts_dir)):
        mkdir(pj(rep, pkglts_dir))

    LOGGER.info("init package")

    for name in ("regenerate.no", "clean.no"):
        if not exists(pj(rep, pkglts_dir, name)):
            with open(pj(rep, pkglts_dir, name), 'w') as fhw:
                fhw.write("")

    if exists(pj(rep, pkglts_dir, pkg_cfg_file)):
        cfg = get_pkg_config(rep)
    else:
        cfg = Config(DEFAULT_CFG)
    write_pkg_config(cfg, rep)

    if not exists(pj(rep, pkglts_dir, pkg_hash_file)):
        write_pkg_hash({}, rep)


def clean(rep="."):
    """Thorough cleaning of all arborescence rooting at rep.

    Todo: exception list instead of hardcoded one

    Args:
        rep (str): default ".", top directory to clean

    Returns:
        None
    """
    for name in ("build", "dist", "doc/_dvlpt", "doc/build"):
        pth = normpath(pj(rep, name))
        if exists(pth):
            rmtree(pth)

    for root, dnames, fnames in walk(rep):
        # do not walk directories starting with "."
        for name in tuple(dnames):
            if "clean.no" in listdir(pj(root, name)):
                dnames.remove(name)
            elif name.startswith("."):
                dnames.remove(name)
            elif name == "__pycache__":
                rmtree(pj(root, name))
                dnames.remove(name)

        for name in fnames:
            if not name.startswith("."):
                if splitext(name)[1] in [".pyc", ".pyo"]:
                    remove(pj(root, name))


def add_option(name, cfg):
    """Add a new option to this package.

    Notes: See the list of available options online

    Args:
        name (str): name of option to add
        cfg (Config):  current package configuration

    Returns:
        (Config): updated package configuration
    """
    if name in cfg.installed_options():
        raise UserWarning("option already included in this package")

    return update_opt(name, cfg)


def install_example_files(option, cfg, target="."):
    """Install example files associated to an option.

    Args:
        option (str): name of option
        cfg (Config):  current package configuration
        target (str): target directory to write into

    Returns:
        (bool): whether operation succeeded or not
    """
    if option not in cfg.installed_options():
        LOGGER.warning("please install option before example files")
        return False

    opt = available_options[option]
    ex_dir = opt.example_dir()
    if ex_dir is None:
        LOGGER.info("option does not provide any example")
        return False

    regenerate_dir(ex_dir, target, cfg, {})
    return True


def _manage_conflicts(target, hm_ref, overwrite):
    """Check if files have been modified.

    Args:
        target (str): target directory to write into
        hm_ref (dict): reference hash for blocks
        overwrite (bool): whether or not, by default,
                         overwrite user modified files

    Returns:
        (dict of str, bool): file path, overwrite
    """
    conflicted = []
    for file_pth in hm_ref:
        pth = pj(target, file_pth)
        if exists(pth) and modified_file_hash(pth, hm_ref):
            conflicted.append(pth)
        else:
            # file disappeared, regenerate_dir will reload it if managed by pkglts
            pass

    overwrite_file = {}
    if conflicted:
        if overwrite:
            for name in conflicted:
                LOGGER.debug("conflicted, '%s'", name)
                overwrite_file[pth_as_key(name)] = True
        else:
            for name in conflicted:
                LOGGER.warning("A non editable section of %s has been modified", name)
                overwrite_file[pth_as_key(name)] = get_user_permission("overwrite", False)

    return overwrite_file


def regenerate_package(cfg, target=".", overwrite=False):
    """Rebuild all automatically generated files.

    Args:
        cfg (Config):  current package configuration
        target (str): target directory to write into
        overwrite (bool): default False, whether or not
                         to overwrite user modified files

    Returns:
        None
    """
    # check consistency of env params
    invalids = []
    for option in cfg.installed_options():
        for name in check_option_parameters(option, cfg):
            invalids.append((option, name))

    if invalids:
        for option, param in invalids:
            LOGGER.warning("param %s is not valid for '%s'", param, option)

        return False

    # check for potential conflicts
    hm_ref = get_pkg_hash(target)
    overwrite_file = _manage_conflicts(target, hm_ref, overwrite)

    # render files for all options
    hmap = {}
    for name in cfg.installed_options():
        opt = available_options[name]
        resource_dir = opt.resource_dir()
        if resource_dir is None:
            LOGGER.info("option %s does not provide files", name)
        else:
            LOGGER.info("rendering option %s", name)
            loc_hm = regenerate_dir(resource_dir, target, cfg, overwrite_file)
            hmap.update(loc_hm)

    hm_ref.update(hmap)
    write_pkg_hash(hm_ref, target)


def regenerate_option(cfg, name, target=".", overwrite=False):
    """Call the regenerate function of a given option

    Args:
        cfg (Config):  current package configuration
        name: (str) name of option
        target: (str) target directory to write into
        overwrite (bool): default False, whether or not
                         to overwrite user modified files

    Returns:
        None
    """
    # test existence of option regenerate module
    try:
        opt = available_options[name]
        opt.regenerate(cfg, target, overwrite)
    except KeyError:
        raise KeyError("option '%s' does not exists" % name)



from os import mkdir, rmdir
from os.path import exists

import pytest
from pkglts.config_management import Config
from pkglts.option.notebook.option import OptionNotebook


@pytest.fixture()
def opt():
    return OptionNotebook('notebook')


@pytest.fixture()
def tmp_dir():
    pth = "nb_test_config"
    mkdir(pth)

    yield pth

    if exists(pth):
        rmdir(pth)


def test_root_dir_is_defined(opt):
    assert opt.root_dir() is not None


def test_update_parameters(opt):
    cfg = {}
    opt.update_parameters(cfg)
    assert len(cfg['notebook']) == 1


def test_config_check_src_directory(opt, tmp_dir):
    cfg = Config(dict(notebook={'src_directory': "failed_nb"}))
    assert 'notebook.src_directory' in opt.check(cfg)

    cfg = Config(dict(notebook={'src_directory': tmp_dir}))
    assert 'notebook.src_directory' not in opt.check(cfg)


def test_require_option(opt):
    assert len(tuple(opt.require_option())) == 1


def test_require(opt):
    cfg = Config()
    opt.update_parameters(cfg)

    assert len(tuple(opt.require(cfg))) == 0



import pytest
from pkglts.config_management import Config
from pkglts.option.pypi.option import OptionPypi


@pytest.fixture()
def opt():
    return OptionPypi('pypi')


def test_root_dir_is_defined(opt):
    assert opt.root_dir() is not None


def test_update_parameters(opt):
    cfg = {}
    opt.update_parameters(cfg)
    assert len(cfg['pypi']) == 2


def test_config_check_classifiers_exists(opt):
    cfg = Config(dict(pypi={'classifiers': []}))
    assert 'pypi.classifiers' in opt.check(cfg)


def test_require_option(opt):
    assert len(tuple(opt.require_option())) == 2


def test_require(opt):
    cfg = Config()
    opt.update_parameters(cfg)

    assert len(tuple(opt.require(cfg))) == 1



"""
Script which can be called before generating the doc
to generate description of files modified by each option.
"""

from os import listdir
from os.path import isdir
from os.path import splitext, join as pj

from pkglts.config_management import Config


def _nn(cfg, pth):
    """Helper function"""
    print("pth:", pth)
    tgt_name = cfg.render(pth)
    if tgt_name.endswith(".tpl"):
        tgt_name = tgt_name[:-4]

    return tgt_name


def _tree(dname, padding, txt):
    """Generate tree ascii representation of a dir"""
    pkg_cfg = dict(base={"namespace": None,
                         "owner": "owner",
                         "pkgname": "pkgname",
                         "url": None},
                   data={"use_ext_dir": False},
                   doc={"fmt": "rst"},
                   plugin_project={"plugin_name": "plugin"},
                   test={'suite_name': "pytest"})

    cfg = Config(pkg_cfg)

    files = [(isdir(pj(dname, fname)), fname) for fname in listdir(dname)]
    files.sort()

    count = 0
    for is_dir, fname in files:
        fmt_name = _nn(cfg, fname)
        if splitext(fmt_name)[0] != '_':
            count += 1
            txt += padding + '|\n'
            txt += padding + '+-' + fmt_name
            path = pj(dname, fname)
            if is_dir:
                txt += "/\n"
                if count == len(files):
                    txt = _tree(path, padding + ' ' + ' ' * int(len(fmt_name) / 2), txt)
                else:
                    txt = _tree(path, padding + '|' + ' ' * int(len(fmt_name) / 2), txt)
            else:
                txt += '\n'

    return txt


def fmt_tree(dname):
    """Generate tree ascii representation of a dir"""
    return _tree(dname, '', ".\n")



"""
This tool will try to parse all release tags to create an history of package.
"""
import logging
import os
from functools import cmp_to_key
from getpass import getpass

try:  # python3
    from urllib.parse import quote_plus
except ImportError:  # python2
    from urllib import quote_plus

import requests
import semver

LOGGER = logging.getLogger(__name__)


def github_tag_list(project):
    """Retrieve tag list from project.

    Notes: release tag format
        a release tag is of the form vX.X.X where
        X.X.X stands for package version

    Args:
        project (str): name of project (e.g. "revesansparole/pkglts")

    Returns:
        (dict of (str|dict)): tag name, tag info
    """
    base_url = "https://api.github.com"
    url = base_url + "/repos/%s/releases" % project
    res = requests.get(url)
    LOGGER.info("status: %s", res.status_code)
    tags = {}
    for tag in res.json():
        if tag['tag_name'].startswith("v"):
            tags[tag['tag_name'][1:]] = dict(name=tag['tag_name'],
                                             date=tag['created_at'][:10],
                                             title=tag['name'],
                                             body=tag['body'])

    return tags


def gitlab_tag_list(server, project, token):
    """Retrieve tag list from project.

    Notes: release tag format
        a release tag is of the form vX.X.X where
        X.X.X stands for package version

    Args:
        server (str): base url of gitlab server (e.g. "framagit.org")
        project (str): name of project (e.g. "agro/agrosim")
        token (str): valid token to access project repo

    Returns:
        (dict of (str|dict)): tag name, tag info
    """
    base_url = "https://%s/api/v4/projects/" % server
    repo_id = quote_plus(project)
    LOGGER.info("repo_id: %s", repo_id)

    url = base_url + repo_id + "/repository/tags" + "?private_token=%s" % token
    LOGGER.debug("url:  %s", url)

    res = requests.get(url)
    LOGGER.debug("status: %s", res.status_code)
    tags = {}
    for tag in res.json():
        if tag['name'].startswith("v"):
            tags[tag['name'][1:]] = dict(name=tag['name'],
                                         date=tag['commit']['committed_date'][:10],
                                         title=tag['message'],
                                         body=tag['release']['description'])

    return tags


def write_changelog(tags, fmt):
    """Write changelog file according to the content of release tags.

    Notes: release tag format
        a release tag is of the form vX.X.X where
        X.X.X stands for package version

    Warnings: invalid tags
        The process will not check that tags are properly formatted.

    Args:
        tags (dict): as returned by one of tag_list function
        fmt (str): format for documentation 'md' or 'rst'

    Returns:
        (None)
    """
    ver_list = list(tags.keys())
    ver_list.sort(key=cmp_to_key(semver.compare), reverse=True)

    # format changelog
    if fmt == "md":
        txt = "# History\n\n"
        for ver in ver_list:
            tag = tags[ver]
            txt += "## %s - <small>*(%s)*</small> - %s\n\n" % (tag['name'], tag['date'], tag['title'])
            txt += tag['body']
            txt += "\n\n"

        for name in ("CHANGELOG.md", "HISTORY.md"):
            if os.path.exists(name):
                LOGGER.info("write changelog in %s", name)
                with open(name, 'w') as fhw:
                    fhw.write(txt)

    elif fmt == "rst":
        txt = "=======\nHistory\n=======\n\n"
        for ver in ver_list:
            tag = tags[ver]
            tag_title = "%s - *(%s)* - %s" % (tag['name'], tag['date'], tag['title'])
            txt += tag_title + "\n"
            txt += "=" * len(tag_title) + "\n\n"
            txt += tag['body']
            txt += "\n\n"

        for name in ("CHANGELOG.rst", "HISTORY.rst"):
            if os.path.exists(name):
                LOGGER.info("write changelog in %s", name)
                with open(name, 'w') as fhw:
                    fhw.write(txt)
    else:
        LOGGER.warning("Doc format '%s' unsupported", fmt)


def action_history(cfg, **kwds):
    """Regenerate history file from tag list.
    """
    LOGGER.info("Reconstruct history")

    # extract release tags from versioning system
    if 'git' not in cfg.installed_options():
        LOGGER.warning("Project is not under git versioning, unable to continue")
        return

    if 'gitlab' in cfg.installed_options():
        server = cfg['gitlab']['server']
        owner = cfg['gitlab']['owner']
        project = cfg['gitlab']['project']
        token = getpass("gitlab API access token:")
        tags = gitlab_tag_list(server, "%s/%s" % (owner, project), token)
    elif 'github' in cfg.installed_options():
        owner = cfg['github']['owner']
        project = cfg['github']['project']
        tags = github_tag_list("%s/%s" % (owner, project))
    else:
        LOGGER.info("git only option not supported yet")
        tags = {}

    # format tags into history file
    if tags:
        write_changelog(tags, cfg['doc']['fmt'])


def parser_history(subparsers):
    """Associate a CLI to this tool.

    Notes: The CLI will be a subcommand of pmg.

    Args:
        subparsers (ArgumentParser): entity to create a subparsers

    Returns:
        (string): a unique id for this parser
        (callable): the action to perform
    """
    parser = subparsers.add_parser('history', help=action_history.__doc__)

    return 'history', action_history



"""
Define actions that can be called with the CLI.
"""
import json
import logging
from argparse import ArgumentParser

from . import logging_tools
from .config_management import get_pkg_config, write_pkg_config
from .manage import add_option, clean, init_pkg, install_example_files, regenerate_option, regenerate_package
from .option_tools import available_options
from .version_management import outdated_options, write_pkg_version, FileNotFoundError


LOGGER = logging.getLogger(__name__)


def action_info(cfg, **kwds):
    """Display info on package for debug purpose.
    """
    LOGGER.info("package info")
    print("current config (after resolution)")
    opt_names = kwds['option']
    all_opts = False
    if not opt_names:
        opt_names = cfg.installed_options()
        all_opts = True

    outdated = outdated_options(cfg)

    for name in opt_names:
        if name in outdated:
            print("%s: OUTDATED" % name)
        else:
            print(name)
        print(json.dumps(cfg[name], sort_keys=True, indent=2))

    if all_opts:
        print("other available options:")
        for opt_name in sorted(set(available_options) - set(cfg.installed_options())):
            print("  ", opt_name)


def action_clean(cfg, **kwds):
    """Clean package of all un necessary files.
    """
    del cfg, kwds  # unused
    LOGGER.info("clean package")
    clean()


def action_init(cfg, **kwds):
    """Initialize environment for use of pkglts.
    """
    if cfg is not None:
        LOGGER.warning("Directory is already a pkglts package")
        return

    init_pkg()

    if kwds:
        action_add(get_pkg_config(), **kwds)


def action_clear(cfg, **kwds):
    """Attempt to free the package from pkglts interactions.
    """
    del cfg, kwds  # unused
    LOGGER.info("clear")
    print("TODO")


def action_update(cfg, **kwds):
    """Check if a new version of pkglts is available.
    """
    del cfg, kwds  # unused
    LOGGER.info("update")
    print("TODO")


def action_regenerate(cfg, **kwds):
    """Regenerate all files in the package.
    """
    outdated = outdated_options(cfg)
    if len(outdated) > 0:
        out_fmt = "\n".join(outdated)
        LOGGER.warning("Some options are outdated,"
                       " please upgrade pkglts and/or all the following options:\n"
                       "%s" % out_fmt)
        return

    clean()

    if kwds['option']:
        for name in kwds['option']:
            LOGGER.info("regenerate '%s'", name)
            regenerate_option(cfg, name, overwrite=kwds['overwrite'])
    else:
        LOGGER.info("regenerate package")
        regenerate_package(cfg, overwrite=kwds['overwrite'])

    write_pkg_version(cfg)


def action_add(cfg, **kwds):
    """Add new options in the package.
    """
    LOGGER.info("add option")
    for name in kwds['option']:
        cfg = add_option(name, cfg)

    write_pkg_config(cfg)


def action_remove(cfg, **kwds):
    """Remove options from the package.
    """
    LOGGER.info("remove option")
    print("TODO")
    del cfg, kwds


def action_example(cfg, **kwds):
    """Install example files associated with options.
    """
    for name in kwds['option']:
        install_example_files(name, cfg)


def main():
    """Run CLI evaluation"""
    action = dict(
        info=action_info,
        clean=action_clean,
        init=action_init,
        clear=action_clear,
        update=action_update,
        rg=action_regenerate,
        add=action_add,
        remove=action_remove,
        example=action_example
    )
    # parse argument line
    parser = ArgumentParser(description='Package structure manager')
    parser.add_argument("-v", "--verbosity", action="count", default=0,
                        help="increase output verbosity")

    subparsers = parser.add_subparsers(dest='subcmd', help='sub-command help')

    parser_info = subparsers.add_parser('info', help=action_info.__doc__)
    parser_info.add_argument('option', nargs='*',
                             help="name of option to fetch info")

    parser_clean = subparsers.add_parser('clean', help=action_clean.__doc__)

    parser_init = subparsers.add_parser('init', help=action_init.__doc__)
    parser_init.add_argument('option', nargs='*',
                             help="name of option to add during init")
    parser_clear = subparsers.add_parser('clear', help=action_clear.__doc__)
    parser_update = subparsers.add_parser('update', help=action_update.__doc__)
    parser_rg = subparsers.add_parser('rg', help=action_regenerate.__doc__)
    parser_rg.add_argument('option', nargs='*',
                           help="name of option to add")
    parser_rg.add_argument("--overwrite", action='store_true',
                           help="Globally overwrite user modified files")

    parser_add = subparsers.add_parser('add', help=action_add.__doc__)
    parser_add.add_argument('option', nargs='+',
                            help="name of option to add")

    parser_remove = subparsers.add_parser('remove', help=action_remove.__doc__)
    parser_remove.add_argument('option', nargs='+',
                               help="name of option to remove")
    parser_example = subparsers.add_parser('example', help=action_example.__doc__)
    parser_example.add_argument('option', nargs='+',
                                help="name of option which offer example files")

    # try to read package config for extra commands
    try:
        cfg = get_pkg_config()
    except FileNotFoundError:
        cfg = None
    else:
        # add option commands
        for opt_name in cfg.installed_options():
            for parser_tool in available_options[opt_name].tools(cfg):
                name, action_tool = parser_tool(subparsers)
                action[name] = action_tool

    kwds = vars(parser.parse_args())
    logging_tools.main(kwds.pop('verbosity'))

    # perform action
    subcmd = kwds.pop('subcmd')
    if cfg is None and subcmd != "init":
        LOGGER.error("Directory is not a pkglts package, run pmg init first")
        return

    action[subcmd](cfg, **kwds)


if __name__ == '__main__':
    main()



"""
This tool will display all dependencies used by this package.
"""
import ast
import logging
from glob import glob
from os import path

from pkglts.local import src_dir

LOGGER = logging.getLogger(__name__)

stdpkgs = set(line.strip() for line in open(path.join(path.dirname(__file__), "stdpkgs.txt")).readlines()
              if len(line.strip()) > 0 and not line.startswith("#"))


def iter_ext_imports(body):
    for node in body:
        if isinstance(node, ast.Import):
            for alias in node.names:
                yield alias.name
        elif isinstance(node, ast.ImportFrom):
            if node.level == 0:  # only external imports
                yield node.module

        # elif isinstance(node, ast.FunctionDef):
        #     for name, level in iter_imports(node.body):
        #         yield name, level


def action_find_reqs(cfg, **kwds):
    """Find dependencies used by this package.
    """
    LOGGER.info("Find requirements")
    this_pkgname = cfg['base']['pkgname']

    for dirpth in (src_dir(cfg), 'doc', 'example', 'test'):
        if path.exists(dirpth):
            print("############\n# %s\n############" % dirpth)
            reqs = set()
            for pth in glob("%s/*.py" % dirpth) + glob("%s/**/*.py" % dirpth):
                if not path.basename(pth).startswith("_"):
                    src = open(pth, 'rb').read()
                    pt = ast.parse(src, pth)
                    # TODO pb with namespaces
                    reqs.update(pkgname.split(".")[0] for pkgname in iter_ext_imports(pt.body))
            reqs -= {this_pkgname}
            print("standard", sorted(reqs & stdpkgs))
            print("external", sorted(reqs - stdpkgs))
            print("")


def parser_find_reqs(subparsers):
    """Associate a CLI to this tool.

    Notes: The CLI will be a subcommand of pmg.

    Args:
        subparsers (ArgumentParser): entity to create a subparsers

    Returns:
        (string): a unique id for this parser
        (callable): the action to perform
    """
    parser = subparsers.add_parser('reqs', help=action_find_reqs.__doc__)

    return 'reqs', action_find_reqs



"""Pip front end to install package in current environment.
"""
import imp
import pip
## Dealing with pip 10.* api changes
## The solution is from:
## https://github.com/naiquevin/pipdeptree/blob/master/pipdeptree.py
try:
    from pip._internal import get_installed_distributions
except ImportError:
    from pip import get_installed_distributions
try:
    from pip._internal import main as pip_install
except:
    from pip import main as pip_install


def installed_packages():
    """Iterate on all packages installed in the current python environment.

    return:
        (iter of str)
    """
    pip.utils.pkg_resources = imp.reload(pip.utils.pkg_resources)
    for pjt in get_installed_distributions():
        yield pjt.project_name


def install(name):
    """Install a package in the current python environment.

    arg:
     - name (str): name of the package

    return:
     - (bool): whether installation was successful or not
    """
    pip_install(['install', name])



from os.path import dirname

from pkglts.dependency import Dependency
from pkglts.option_object import Option
from pkglts.version import __version__


def is_valid_identifier(name):
    """ Check that name is a valid python identifier
    sort of back port of "".isidentifier()
    """
    try:
        compile("%s=1" % name, "test", 'single')
        return True
    except SyntaxError:
        return False


class OptionPluginProject(Option):
    def version(self):
        return __version__

    def root_dir(self):
        return dirname(__file__)

    def update_parameters(self, cfg):
        cfg['plugin_project'] = {
            "plugin_name": "{{ base.pkgname }}"
        }

    def check(self, cfg):
        invalids = []
        plugin_name = cfg['plugin_project']['plugin_name']

        if "." in plugin_name:
            invalids.append('plugin_project.plugin_name')
        elif not is_valid_identifier(plugin_name):
            invalids.append('plugin_project.plugin_name')

        return invalids

    def require_option(self):
        return ['pysetup', 'data', 'git']

    def require(self, cfg):
        del cfg
        yield Dependency('pkglts', intent='install', channel='revesansparole')



from os.path import abspath, basename, dirname

from pkglts.local import pkg_full_name, src_dir
from pkglts.option_object import Option
from pkglts.version import __version__


def is_valid_identifier(name):
    """ Check that name is a valid python identifier
    sort of back port of "".isidentifier()
    """
    try:
        compile("%s=1" % name, "test", 'single')
        return True
    except SyntaxError:
        return False


class OptionBase(Option):
    def version(self):
        return __version__

    def root_dir(self):
        return dirname(__file__)

    def update_parameters(self, cfg):
        sec = dict(
            pkgname=basename(abspath(".")),
            namespace=None,
            namespace_method="pkg_util",
            url=None,
            authors=[("moi", "moi@email.com")]
        )
        cfg['base'] = sec

    def check(self, cfg):
        invalids = []
        pkgname = cfg['base']['pkgname']
        namespace = cfg['base']['namespace']

        if "." in pkgname:
            invalids.append('base.pkgname')
        elif not is_valid_identifier(pkgname):
            invalids.append('base.pkgname')

        if namespace is not None:
            if "." in namespace:
                invalids.append('base.namespace')
            elif not is_valid_identifier(namespace):
                invalids.append('base.namespace')

        if cfg['base']['namespace_method'] not in ("pkg_util", "setuptools", "P3.3>"):
            invalids.append("base.namespace_method")

        return invalids

    def environment_extensions(self, cfg):
        return {"pkg_full_name": pkg_full_name(cfg),
                "src_pth": src_dir(cfg)}



"""
Set of function used to regenerate html files from notebooks.
"""
import fnmatch
import os
import os.path
import shutil


def write_rst_file_with_resources(body, resources):
    """Helper function.
    """
    import nbconvert

    # Create folder if the path not exists
    if not os.path.exists(resources["metadata"]["path"]):
        os.makedirs(resources["metadata"]["path"])

    # Keep resources metadata like image in the rst
    writer = nbconvert.writers.FilesWriter()
    # TODO hack to solve problems with pygment not recognizing ipython2
    body = body.replace(".. code:: ipython2", ".. code:: python")
    body = body.replace(".. code:: ipython3", ".. code:: python")
    writer.write(body, resources, notebook_name=resources["metadata"]["name"])


def write_rst_file_with_filename(body, filename):
    """
    Warning, This write not save image in the transformation to rst format.
    """
    # Create folder if the path not exists
    if not os.path.exists(os.path.dirname(filename)):
        os.makedirs(os.path.dirname(filename))

    # Write rst file
    with open(filename, "w+") as rst_file:
        rst_file.write(body)


def find_notebook_file(root_directory):
    """Get recursively all notebook filenames in the src_directory.

    Args:
        root_directory (str): base directory to search

    Returns:
        (list of str)
    """
    matches = []
    for root, _, filenames in os.walk(root_directory):
        for filename in fnmatch.filter(filenames, '*.ipynb'):

            # Avoid tmp notebook in hidden folder like *-checkpoint.ipynb
            if os.path.basename(root):
                if os.path.basename(root)[0] == '.':
                    continue

            matches.append(os.path.join(root, filename))

    return matches


def action_nbcompile(cfg, **kwds):
    """Compile notebooks into html.
    """
    import nbconvert

    src_directory = cfg["notebook"]['src_directory']
    len_src_directory = len(src_directory)

    nb_filenames = find_notebook_file(src_directory)

    # define documentation rst notebook directory
    dst_rst_directory = os.path.join("doc", "_notebook")

    # remove previous folder
    if os.path.exists(dst_rst_directory):
        shutil.rmtree(dst_rst_directory)

    index_body = """\
========
Notebook
========

.. toctree::
    :glob:
    :caption: Notebook

"""

    rst_exporter = nbconvert.RSTExporter()
    for nb_filename in nb_filenames:
        # Convert each notebook to rst
        body, resources = rst_exporter.from_filename(nb_filename)

        # Remove basename src_directory in the path
        resources["metadata"]["path"] = resources["metadata"]["path"][len_src_directory + 1:]

        # Get the local file_path of the notebook write
        local_file_path = os.path.join(resources["metadata"]["path"], resources["metadata"]["name"])

        # Add dst_rst_directory in the path
        resources["metadata"]["path"] = os.path.join(dst_rst_directory, resources["metadata"]["path"])

        # Write beginning block to download file
        disclaimer = "This file has been generated from the following notebook: :download:`%s`.\n\n" % os.path.basename(nb_filename)
        disclaimer += "Download it if you want to replay it using `jupyter notebook <http://jupyter.org/>`_.\n\n"

        body = disclaimer + body

        # Write rst with his resources
        write_rst_file_with_resources(body, resources)

        # Write notebook file for further download
        shutil.copy(nb_filename, os.path.join(dst_rst_directory, os.path.basename(nb_filename)))

        # Save the notebook rst position in the index body
        index_body += "    " + local_file_path.replace("\\", "/") + "\n"

    # Write rst_index body in the root src directory
    rst_index_filename = os.path.join(dst_rst_directory, "index.rst")
    write_rst_file_with_filename(index_body, rst_index_filename)


def parser_nbcompile(subparsers):
    """Associate a CLI to this tool.

    Notes: The CLI will be a subcommand of pmg.

    Args:
        subparsers (ArgumentParser): entity to create a subparsers

    Returns:
        (string): a unique id for this parser
        (callable): the action to perform
    """
    parser = subparsers.add_parser('nbcompile', help=action_nbcompile.__doc__)

    return 'nbcompile', action_nbcompile
